/** Classe AbulEduApplicationV1
  *
  * Classe de base en version une pour les applications AbulEdu permettant de structurer
  * nos logiciels et respecter ainsi un certain nombre de choses dont l'interface publique
  * à laquelle le plugin peut se connecter pour échanger des données (logs pour le moment)
  *
  * @see https://redmine.ryxeo.com/projects/ryxeo/wiki/AbulEduApplicationV1
  * @author 2009-2012 Eric Seigne <eric.seigne@ryxeo.com>
  * @author 2011 Jean-Louis Frucot <frucot.jeanlouis@free.fr>
  * @see The GNU Public License (GPL)
  *
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
  * for more details.
  *
  * You should have received a copy of the GNU General Public License along
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  */

#include "abuleduapplicationv1.h"

/** constructeur ... */
AbulEduApplicationV1::AbulEduApplicationV1(int &argc, char *argv[], QString appName, QString appVersion, QString organizationDomain, QString organizationName) :
    QtSingleApplication(argc, argv),
    m_sharedMemory("abeUpdateRequest")
{
    m_localDebug = 0;
    if(m_localDebug)
        qDebug() << "AbulEduApplicationV1: constructeur";
    setApplicationName(appName);
    setApplicationVersion(appVersion);
    setOrganizationName(organizationName);
    setOrganizationDomain(organizationDomain);

    if(organizationName == "")
        qWarning() << "L'appel au constructeur AbulEduApplicationV1 ne specifie pas d'organizationName ... le chargement de plugins personnels ne sera pas automatique.";

    //Initialisation des différentes variables
    m_applicationName  = "";
    m_decorRatio       = 1;
    int already_requested = 0;

    m_nam = new AbulEduNetworkAccessManagerV1(this);

    //pas la peine d'aller plus loin si le developpeur n'a pas fait son boulot !
    if(applicationName() == "" || applicationVersion() == "") {
        qWarning() << trUtf8("AbulEduApplicationV1:: Attention: applicationName ou applicationVersion non défini !!!");
        qWarning() << applicationName() << "::" << applicationVersion();
        //on essaye tout de meme de charger les plugins
        loadPlugins();
        return;
    }
    // ============================================================================== memoire partagee et requete reseau update
    if (m_sharedMemory.isAttached())
        detach();
    //creation du segment de memoire partagee
    if(m_sharedMemory.create(sizeof(already_requested))) {
        if(m_localDebug)
            qDebug() << "AbulEduApplicationV1 : creation de la zone memoire partagee, taille" << m_sharedMemory.size();

        // ----------------------------------
        //On cherche pour voir s'il y a une mise a jour de l'application de disponible
        //On lance notre requete reseau et on gere la suite

        //Il faudrait pouvoir lancer plusieurs requetes en // sur differents supports
        //avec ou sans proxy, en http et en https, voir meme sur d'autres domaines
        QUrl urlUpdates("http://updates.ryxeo.com/application/" + applicationName() + "/version/" + applicationVersion() + "/os/" + os() + "/dist/" + linuxDistrib() + "/format/xml/uid/" + uniqueUID() + "/age/" + QString::number(execAge()).toAscii().toBase64() + "/");
        connect(m_nam, SIGNAL(finished(QNetworkReply*)),this, SLOT(slotOnlineUpdateRequest(QNetworkReply*)));
        m_networkReply = m_nam->get(QNetworkRequest(urlUpdates));
        m_networkReply->setObjectName("OnlineUpdateRequest");
        connect(m_networkReply, SIGNAL(error(QNetworkReply::NetworkError)), this, SLOT(slotReplyError(QNetworkReply::NetworkError)));
        // ----------------------------------


        // on écrit le flag comme quoi la requete reseau est deja partie ... merci de ne pas relancer cette requete :)
        m_sharedMemory.lock();
        already_requested = 1;
        if(m_localDebug)
            qDebug() << "AbulEduApplicationV1 : ecriture dans la memoire partagee: requete reseau lancee";
        memcpy(m_sharedMemory.data(), &already_requested, sizeof(already_requested));
        m_sharedMemory.unlock();
    }
    else {
        if(m_localDebug)
            qDebug() << "AbulEduApplicationV1 : zone memoire partagee deja existante, la requete reseau a deja ete lancee";
        // si le segment existe deja on l'utilise
        if(m_sharedMemory.error() == QSharedMemory::AlreadyExists) {
            if(m_sharedMemory.attach()) {
                // lecture des donnees
                m_sharedMemory.lock();
                already_requested = *(bool *)m_sharedMemory.constData();
                m_sharedMemory.unlock();
            }
            else {
                if(m_localDebug)
                    qDebug() << "AbulEduApplicationV1 : erreur sur sharedMemory :: " << m_sharedMemory.errorString();
            }
        }
        else {
            if(m_localDebug)
                qDebug() << "AbulEduApplicationV1 : erreur sur sharedMemory :: " << m_sharedMemory.errorString();
        }
    }
    // ============================================================================== fin de gestion memoire partagee et requete reseau update


    // ============================================================================== logiciel de plus de six mois
    QFileInfo fi(qApp->applicationFilePath());
    QDateTime ageDuFichier(fi.created());
    QDateTime agePlusSixMois(ageDuFichier.addMonths(24));
    QUrl urlSixMois("http://updates.ryxeo.com/application/" + applicationName() + "/version/" + applicationVersion() + "/os/" + os() + "/dist/" + linuxDistrib() + "/format/xml/uid/" + uniqueUID() + "/age/" + QString::number(execAge()).toAscii().toBase64() + "/");
    //Si l'application a plus de six mois ... on ouvre un browser sur la page speciale
    //pour s'assurer que des très vieux logiciels ne sont plus utilisés
    //passé a 24 mois, on reviendra sur l'idée de 6 mois mais ça fait quand meme un peu court 6 mois, 2 ans c'est déjà
    //un peu plus compréhensible
    if(agePlusSixMois < QDateTime::currentDateTime()) {
        AbulEduNetworkThreadV1 *browser = new AbulEduNetworkThreadV1(urlSixMois);
        browser->start();
        m_browserLaunched = true;
    }
    // ============================================================================== fin logiciel de plus de six mois

    // ============================================================================== plus de x lancements sans reponse de serveur

    // ============================================================================== fin plus de x lancements sans reponse de serveur
    if(m_localDebug)
        qDebug() << "AbulEduApplicationV1 : avant loadPlugins";
    loadPlugins();
    if(m_localDebug)
        qDebug() << "AbulEduApplicationV1 : après loadPlugins";
}

AbulEduApplicationV1::~AbulEduApplicationV1()
{
    if(m_localDebug)
        qDebug() << "AbulEduApplicationV1: destructeur";

    //On detache la memoire partagee
    detach();
}

void AbulEduApplicationV1::loadPlugins()
{
    // ---------------------------------- chargement plugin
    //Ensuite on charge les plugins qui se trouvent dans le sous dossier plugins de l'application
    QDir pluginsDir(applicationDirPath());
    if (pluginsDir.dirName().toLower() == "debug" || pluginsDir.dirName().toLower() == "release")
        pluginsDir.cdUp();
#if defined(Q_OS_MAC)
    if (pluginsDir.dirName() == "MacOS") {
        pluginsDir.cdUp();
        pluginsDir.cdUp();
        pluginsDir.cdUp();
    }
#endif
    if(pluginsDir.cd("plugins")) {
        foreach (QString fileName, pluginsDir.entryList(QDir::Files)) {
            if(m_localDebug)
                qDebug() << "AbulEduApplicationV1 : Chargement du plugin " << fileName;
            QPluginLoader pluginLoader(pluginsDir.absoluteFilePath(fileName));
            QObject *plugin = pluginLoader.instance();
            if(m_localDebug)
                qDebug() << "AbulEduApplicationV1 : Chargement du plugin (2)" << plugin;
            if (plugin && m_localDebug) {
                qDebug() << "AbulEduApplicationV1 : Chargement du plugin " << fileName << " réussi";
            }
        }
    }
    // ---------------------------------- Fin du chargement plugin

    // ---------------------------------- chargement plugin
    //Puis les plugins qui se trouve dans le repertoire de l'utilisateur (permet de mises a jours "douces")
    QString pluginsDirPerso(QDir::homePath() + "/" + organizationName() + "/" + applicationName() + "/plugins");
    if(m_localDebug)
        qDebug() << "AbulEduApplicationV1 : Chargement des plugins perso " << pluginsDirPerso;
    if(pluginsDir.cd(pluginsDirPerso)) {
        foreach (QString fileName, pluginsDir.entryList(QDir::Files)) {
            if(m_localDebug)
                qDebug() << "AbulEduApplicationV1 : Chargement du plugin personnel " << fileName;
            QPluginLoader pluginLoader(pluginsDir.absoluteFilePath(fileName));
            QObject *plugin = pluginLoader.instance();
            if(m_localDebug)
                qDebug() << "AbulEduApplicationV1 : Chargement du plugin personnel (2)" << plugin;
            if (plugin && m_localDebug) {
                qDebug() << "AbulEduApplicationV1 : Chargement du plugin personnel " << fileName << " réussi";
            }
        }
    }
    // ---------------------------------- Fin du chargement plugin
}

void AbulEduApplicationV1::linuxAnalyse()

{
    //Signature des fedora
    QFile fedora("/etc/fedora-release");
    //Pour l'instant c'est uniquement ubuntu qui propose ca
    QFile lsb("/etc/lsb-release");
    if(lsb.exists()) {
        if (lsb.open(QIODevice::ReadOnly | QIODevice::Text)) {
            QRegExp release("DISTRIB_CODENAME=(.*)");
            while (!lsb.atEnd() && m_linuxDistrib == "") {
                QByteArray line = lsb.readLine();
                release.indexIn(line);
                if(release.cap(1).trimmed() != "") {
                    if(m_localDebug)
                        qDebug() << "AbulEduApplicationV1::linuxPackageFormat On a trouve le LSB :: " << line << " ici : " << release.cap(1).trimmed();
                    m_linuxDistrib = release.cap(1).trimmed();
                    m_linuxPackageFormat = "deb";
                }
            }
            lsb.close();
        }
    }
    else if(fedora.exists()) {
        m_linuxPackageFormat = "rpm";
    }
    else {
        if(m_localDebug)
            qDebug() << "Lancer la commande 'lsb_release -a";
    }
}

QString AbulEduApplicationV1::os()
{
    QString os("");
#if defined(Q_OS_MAC)
    os="osx";
#elif defined(Q_OS_WIN32)
    os="windows";
#elif defined(Q_OS_LINUX)
    QString arch(""); //pour l'instant i386 / amd64
    //L'architecture ... pour l'instant on ne fait pas les ARM / Sparc etc.
#ifdef __x86_64__
    arch = "amd64";
#else
    arch = "i386";
#endif

    //On regroupe tout ca linux-deb-lucid-amd64
    os="linux-" + linuxDistrib() + "-" + linuxPackageFormat() + "-" + arch + "-";
#endif
    return os;
}

QString AbulEduApplicationV1::uniqueUID()
{
    QByteArray retour("000102030405");
    foreach(QNetworkInterface interface, QNetworkInterface::allInterfaces()) {
        if (interface.flags().testFlag(QNetworkInterface::IsUp) && !interface.flags().testFlag(QNetworkInterface::IsLoopBack)) {
            foreach (QNetworkAddressEntry entry, interface.addressEntries()) {
                if ( interface.hardwareAddress() != "00:00:00:00:00:00" && entry.ip().toString().contains(".")) {
                    retour.clear();
                    retour.append(interface.hardwareAddress().remove(":").toLower());
                }
            }
        }
    }
    return retour.toHex().toBase64();
}

int AbulEduApplicationV1::execAge()
{
    QFileInfo fi(qApp->applicationFilePath());
    int ageDuFichier(fi.created().daysTo(QDateTime::currentDateTime()));
    if(m_localDebug)
        qDebug() << "Age du fichier :: " << ageDuFichier;
    return ageDuFichier;
}

QString AbulEduApplicationV1::linuxDistrib()
{
    if(m_linuxDistrib == "")
        linuxAnalyse();
    return m_linuxDistrib;
}

QString AbulEduApplicationV1::linuxPackageFormat()
{
    if(m_linuxPackageFormat == "")
        linuxAnalyse();
    return m_linuxPackageFormat;
}

void AbulEduApplicationV1::setAbeApplicationDebugLevel(int i)
{
    m_localDebug = i;
}

void AbulEduApplicationV1::setAbeApplicationLongName(QString s)
{
    m_applicationName = s;
}

QString AbulEduApplicationV1::getAbeApplicationLongName()
{
    return(m_applicationName);
}

void AbulEduApplicationV1::slotOnlineUpdateRequest(QNetworkReply *reply)
{

    if(m_localDebug)
        qDebug() << "AbulEduApplicationV1 : slotOnlineUpdateRequest " << reply->objectName();

    //Si ce n'est pas pour moi je droppe
    if(reply->objectName() != "OnlineUpdateRequest") {
        return;
    }

    disconnect(m_nam, SIGNAL(finished(QNetworkReply*)),this, SLOT(slotOnlineUpdateRequest(QNetworkReply*)));

    // Pas d'erreur ?
    if (reply->error() == QNetworkReply::NoError) {
        int v = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();
        if (v >= 200 && v < 300) // Success
        {
            QString titre("");
            QString desc("");
            QString link("");
            QString lang("");
            QDomDocument *xml = new QDomDocument();
            xml->setContent(reply->readAll());
            if(m_localDebug)
                qDebug() << "AbulEduApplicationV1 : download XML update " << xml->toString();

            QDomElement xmlElem = xml->documentElement();
            QDomNode node = xmlElem.firstChild();
            while(!node.isNull()) {
                QDomElement element = node.toElement(); // On essaie de convertir le node en element
                if(!element.isNull()) {
                    // Le node est bien un élément.
                    if(element.tagName()=="application") {
                        QDomNode node=element.firstChild(); //On parcourt le document à partir du tag "application"
                        while(!node.isNull()) {
                            QDomElement elem=node.toElement(); //On cherche le titre du "Channel"
                            if(elem.tagName()=="title") {
                                titre = elem.text();
                            }
                            else if((elem.tagName()=="link")) {
                                link = elem.text();
                            }
                            else if((elem.tagName()=="description")) {
                                desc = elem.text();
                            }
                            else if((elem.tagName()=="language")) {
                                lang = elem.text();
                            }
                            node=node.nextSibling();
                        }
                    }
                    node = node.nextSibling();
                }
            }

            //On affiche une boite d'information (popup) à l'utilisateur ...
            if(titre.length()) {
                QMessageBox msgBox;
                msgBox.setWindowTitle(titre);
                msgBox.setText(desc);
                msgBox.setStandardButtons(QMessageBox::Ok);
                msgBox.exec();
            }
            delete xml;
        }
        //S'il s'agit de redirections ...
        else if (v >= 300 && v < 400)
        {
            // Get the redirection url
            QUrl newUrl = reply->attribute(QNetworkRequest::RedirectionTargetAttribute).toUrl();
            // Because the redirection url can be relative,
            // we have to use the previous one to resolve it
            newUrl = reply->url().resolved(newUrl);
            QNetworkRequest redirection(newUrl);
            m_networkReply = m_nam->get(redirection);
            return; // to keep the manager for the next request
        }
    }
}

void AbulEduApplicationV1::slotReplyError(QNetworkReply::NetworkError code)
{
    if(m_localDebug)
        qDebug() << "AbulEduApplicationV1::slotReplyError " << code;
}

void AbulEduApplicationV1::detach()
{
    if (!m_sharedMemory.detach()) {
        if(m_localDebug)
            qDebug() << "AbulEduApplicationV1 : Unable to detach from shared memory.";
    }
}

void AbulEduApplicationV1::setAbeApplicationDecorRatio(float ratio)
{
    m_decorRatio = ratio;
}

float AbulEduApplicationV1::getAbeApplicationDecorRatio()
{
    return m_decorRatio;
}

AbulEduNetworkAccessManagerV1 *AbulEduApplicationV1::getAbeNetworkAccessManager()
{
    return m_nam;
}

void AbulEduApplicationV1::sendSignalDownloaded(QString s)
{
    emit signalDownloaded(s);
}
