/** Classe AbulEduNetworkAccessManagerV1
  * @author 2011-2012 Eric Seigne <eric.seigne@ryxeo.com>
  * @see The GNU Public License (GNU/GPL) v3
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
  * for more details.
  *
  * You should have received a copy of the GNU General Public License along
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  */

#include "abuledunetworkaccessmanagerv1.h"
#include "abuleduapplicationv1.h"
#include "abuledussologinv1.h"

AbulEduNetworkAccessManagerV1::AbulEduNetworkAccessManagerV1(QObject *parent) :
    QNetworkAccessManager(parent)
{
    m_localDebug       = false;
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1: creation";

    m_ssoPopupIsShow   = false;
    m_proxyPassAsked   = 0;
    m_passAsked        = 0;
    m_ssoAuthentifiedStatus = 0;
    m_ssoUsePopup           = true; //par défaut on ouvre une popup si on a besoin d'une authentification SSO
//    m_ssoAuthenticator = QSharedPointer<QAuthenticator>(new QAuthenticator);

    /** Ajout des certificats SSL racine de nos serveurs https (la version de Qt utilisée sous windows ne semble pas avoir ces certificats) */
    QStringList certificats;
    certificats << ":/abuledunetworkaccessmanagerv1/sslrapidssl-01" << ":/abuledunetworkaccessmanagerv1/sslgeotrust-01"
                << ":/abuledunetworkaccessmanagerv1/ssladdtrust-01" << ":/abuledunetworkaccessmanagerv1/sslpositive-01";
    QListIterator<QString> i(certificats);
    while (i.hasNext()) {
        QFile certif(i.next());
        if (certif.open(QIODevice::ReadOnly | QIODevice::Text)) {
            QByteArray c = certif.readAll();
            QSslCertificate cert(c);
            QList<QSslCertificate> certificates = QSslSocket().caCertificates();
            if(!certificates.contains(cert)) {
                if(m_localDebug)
                    qDebug() << " Ajout de " << certif.fileName();
                QSslSocket().addDefaultCaCertificate(cert);
            }
            certif.close();
        }
    }

    //Passage de ticket SSO entre alacarte et les clients
    //Si on est dans un logiciel alors on est client de abuledu-alacarte-srv
    //et on lance un "serveur" sous notre nom
    if(abeApp->applicationName() != "abuledu-alacarte") {
        m_localSocketServer = new AbulEduLocalSocketIpcServerV1(abeApp->applicationName(), this);
        m_localSocketClient = new AbulEduLocalSocketIpcClientV1("abuledu-alacarte-serveur", this);
        connect(m_localSocketServer, SIGNAL(abeSignalMessageReceived(QString)), this, SLOT(slotLocalSocketMessageReceived(QString)),Qt::UniqueConnection);

        //Etn on essaye de demander a alacarte s'il n'a pas les infos SSO
        if(m_login == "" || m_pass == "") {
            //On essaye de demander a alacarte s'il a pas les infos ?
            m_localSocketClient->abeSlotSendMessageToServer("HELLO:" + abeApp->applicationName() + ":ssoRequest:");
        }

    }
    else {
        m_localSocketServer = new AbulEduLocalSocketIpcServerV1("abuledu-alacarte-serveur", this);
        m_localSocketClient = new AbulEduLocalSocketIpcClientV1("vide", this);
        connect(m_localSocketServer, SIGNAL(abeSignalMessageReceived(QString)), this, SLOT(slotLocalSocketMessageReceived(QString)),Qt::UniqueConnection);
    }

    //connexion sur la gestion de l'authentification si c'est necessaire
    connect(this, SIGNAL(proxyAuthenticationRequired(QNetworkProxy,QAuthenticator*)), this, SLOT(slotProxyAuthenticationRequired(QNetworkProxy,QAuthenticator*)),Qt::UniqueConnection);
    //connexion sur la gestion des erreurs LDAP
    connect(this, SIGNAL(sslErrors(QNetworkReply*,QList<QSslError>)), this, SLOT(slotSslErrors(QNetworkReply*,QList<QSslError>)),Qt::UniqueConnection);
    //connexion sur l'authentification SSO
    connect(this,SIGNAL(authenticationRequired(QNetworkReply*,QAuthenticator*)),this,SLOT(slotSSOAuth(QNetworkReply*,QAuthenticator*)), Qt::UniqueConnection);

    //Gestion d'un proxy ?
    //il peut-être déterminé par une variable d'environnement
    QString envproxy = qgetenv("http_proxy");
    if(envproxy.length() > 0)
    {
        if(m_localDebug)
            qDebug() << "AbulEduNetworkAccessManagerV1 Proxy detected :: " << envproxy;
        envproxy.remove("/");
        //On découpe le proxy
        QRegExp exp(":");
        QStringList split = envproxy.split(exp);

        if(m_localDebug)
            qDebug() << "AbulEduNetworkAccessManagerV1 Proxy detected :: " << split;
        QNetworkProxy proxy;
        proxy.setType(QNetworkProxy::HttpProxy);
        proxy.setHostName(split[1]);
        proxy.setPort(split[2].toInt());
        QNetworkProxy::setApplicationProxy(proxy);
        this->setProxy(proxy);
    }
    else {
        //Détection automatique: WPAD
        QUrl adresseWPAD("");
        QHostInfo wpadSingleResolve = QHostInfo::fromName("wpad");
        if (wpadSingleResolve.error() != QHostInfo::NoError) {
            if(m_localDebug)
                qDebug() << "AbulEduNetworkAccessManagerV1 Lookup failed:" << wpadSingleResolve.errorString();
            //On essaye avec le domaine complet ...
            QHostInfo wpadDomainResolve = QHostInfo::fromName("wpad." + QHostInfo::localDomainName());
            if (wpadDomainResolve.error() != QHostInfo::NoError) {
                if(m_localDebug)
                    qDebug() << "AbulEduNetworkAccessManagerV1 Lookup failed:" << wpadDomainResolve.errorString();
            }
            else {
                //On télécharge le fichier wpad pour l'analyser ...
                adresseWPAD = "http://wpad." + QHostInfo::localDomainName() + "/wpad.dat";
            }
        }
        else {
            //On télécharge le fichier wpad pour l'analyser ...
            adresseWPAD = "http://wpad/wpad.dat";
        }
        if(! adresseWPAD.isEmpty()){
            connect(this, SIGNAL(finished(QNetworkReply*)),this, SLOT(finishedSlotWPAD(QNetworkReply*)),Qt::UniqueConnection);
            QNetworkReply* reply = this->get(QNetworkRequest(adresseWPAD));
            reply->setObjectName("WPADRequest");

            //attente active du slot proxy ... attention risque de no limit ?
            QEventLoop loop;
            connect(reply, SIGNAL(finished()), &loop, SLOT(quit()),Qt::UniqueConnection);
            loop.exec();
        }
    }

}

void AbulEduNetworkAccessManagerV1::setAbeLoginPass(QString login, QString pass)
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::setAbeLoginPass (1) : " << login << " " << pass;
    m_login = login;
    m_pass  = pass;
    m_ssoAuthentifiedStatus = 0;
    m_passAsked = 0;
}

void AbulEduNetworkAccessManagerV1::abeSSOLogin()
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::abeSSOLogin (1)";
    //Si jamais login / pass sont vide on affiche la popup ?
    if(m_login == "" || m_pass == "") {
        //evite d'afficher 10 fois la popup si on change de bureau
        if(m_ssoPopupIsShow == false && m_ssoUsePopup == true) {
            AbulEduSSOLoginV1 *a = new AbulEduSSOLoginV1();
            a->show();
            m_ssoPopupIsShow = true;
            connect(a, SIGNAL(destroyed()), this, SLOT(slotSSOPopupClosed()),Qt::UniqueConnection);
            connect(a, SIGNAL(signalClosed()), this, SLOT(slotSSOPopupClosed()),Qt::UniqueConnection);
        }
    }
    QNetworkRequest request;
    QUrl urlLogin("https://webadminxml.abuledu.net/login.php");

    //Grosse amélioration pour gérer les "erreurs de mots de passe", bien plus facile et efficace que de gérer un
    //pointeur sur un QAuthenticator qui se carapate et fait des erreurs de segmentation !
    urlLogin.setPassword(m_pass);
    urlLogin.setUserName(m_login);
    request.setUrl(urlLogin);
    QNetworkReply *reply = this->get(request);
    reply->setObjectName("ssoLogin");
    connect(this, SIGNAL(finished(QNetworkReply*)),this, SLOT(slotFinishedSSOLogin(QNetworkReply*)),Qt::UniqueConnection);
}

void AbulEduNetworkAccessManagerV1::abeSSOLogout()
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::abeSSOLogout (1)";

    m_login = "";
    m_pass = "";
    m_ssoAuthentifiedStatus = -1;
    m_passAsked = 0;

    QNetworkRequest request;
    request.setUrl(QUrl("https://webadminxml.abuledu.net/logout.php"));
    connect(this, SIGNAL(finished(QNetworkReply*)),this, SLOT(slotFinishedSSOLogout(QNetworkReply*)),Qt::UniqueConnection);
    QNetworkReply *reply = this->get(request);
    reply->setObjectName("ssoLogout");
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::abeSSOLogout (2)";
}

int AbulEduNetworkAccessManagerV1::abeSSOAuthenticationStatus()
{
    return m_ssoAuthentifiedStatus;
}

/** @warning: Solution rapide de recherche d'un proxy, je n'analyse pas le javascript retourne et cherche uniquement une ligne PROXY ...*/
void AbulEduNetworkAccessManagerV1::finishedSlotWPAD(QNetworkReply* reply)
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1 finishedSlotWPAD : " << reply->errorString();
    if(reply->objectName() == "WPADRequest") {
        if (reply->error() == QNetworkReply::NoError)
        {
            // lectures des données issues du "reply"
            QByteArray contenu = reply->readAll();
            if(contenu.contains("PROXY")) {
                QByteArray temp = contenu.right(contenu.length() - contenu.indexOf("PROXY ") - 6); //6 = longueur du mot "PROXY "
                QString envproxy = temp.left(temp.indexOf(";"));
                if(m_localDebug)
                    qDebug() << "AbulEduNetworkAccessManagerV1:finishedSlotWPAD Proxy trouvé: " << envproxy;
                QRegExp exp(":");
                QStringList split = envproxy.split(exp);

                if(m_localDebug)
                    qDebug() << "AbulEduNetworkAccessManagerV1:finishedSlotWPAD Proxy detected :: " << split;
                QNetworkProxy proxy;
                proxy.setType(QNetworkProxy::HttpProxy);
                proxy.setHostName(split.at(0));
                proxy.setPort(split.at(1).toInt());
                QNetworkProxy::setApplicationProxy(proxy);
                this->setProxy(proxy);
            }
        }
        disconnect(this,SIGNAL(finished(QNetworkReply*)),this, SLOT(finishedSlotWPAD(QNetworkReply*)));
    }
}

void AbulEduNetworkAccessManagerV1::abeSSOPopupEnable(bool enable)
{
    m_ssoUsePopup = enable;
}

void AbulEduNetworkAccessManagerV1::slotSSOAuth(QNetworkReply *networkReply, QAuthenticator *authenticator)
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::ssoAuth() (1)" << m_passAsked;
    //Si le login / pass est déjà passé c'est que c'est pas bon
    if(m_passAsked > 0 && authenticator->user() == m_login && authenticator->password() == m_pass) {
        m_ssoAuthentifiedStatus = -1;
        if(m_localDebug)
            qDebug() << "Login ou mot de passe incorrect";
    }
    else {
        qDebug() << "  Demande d'authentification " << m_login << " " << m_pass;
        qDebug() << "  Demande d'authentification " << authenticator->user() << " " << authenticator->password();
        authenticator->setUser(m_login);
        authenticator->setPassword(m_pass);
        qDebug() << "  Demande d'authentification " << authenticator->user() << " " << authenticator->password();
        //Detection loop
        //if(m_localDebug)
        m_ssoAuthentifiedStatus = 1;
    }
    m_passAsked++;
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::ssoAuth() (2)";
}

void AbulEduNetworkAccessManagerV1::slotFinishedSSOLogin(QNetworkReply *reply)
{
    if(m_localDebug) {
        qDebug() << "AbulEduNetworkAccessManagerV1::slotFinishedSSOLogin (1) " << reply->objectName();
    }

    if(reply->objectName() == "ssoLogin") {
        qDebug() << "  -> Code retour de reply : " << reply->error();
        if(m_localDebug) {
            qDebug() << "  -> Code retour de reply : " << reply->error();
        }
        if(reply->error() == QNetworkReply::NoError) {
            //Tout va bien
            m_ssoAuthentifiedStatus = 1;
        }
        else if(reply->error() == QNetworkReply::AuthenticationRequiredError) {
            //C'est la qu'on teste si on a fait plusieurs tentatives d'authentification
            if(m_passAsked >= 1) {
                m_ssoAuthentifiedStatus = -1;
                qDebug() << "      authentification erreur -> pourquoi ssoAuth n'est pas lance mystere ?";
                if(m_localDebug) {
                    qDebug() << "      authentification erreur";
                }
            }
            else {
                qDebug() << "      a gerer ici";
                m_ssoAuthentifiedStatus = -1;
            }
        }
        //        else if()
        disconnect(this, SIGNAL(finished(QNetworkReply*)),this, SLOT(slotFinishedSSOLogin(QNetworkReply*)));
    }
    emit ssoAuthStatus(m_ssoAuthentifiedStatus);

    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::slotFinishedSSOLogin (2)";
}

void AbulEduNetworkAccessManagerV1::slotFinishedSSOLogout(QNetworkReply *reply)
{
    if(m_localDebug) {
        qDebug() << "AbulEduNetworkAccessManagerV1::slotFinishedSSOLogout (1)";
        qDebug() << reply->objectName();
    }

    if(reply->objectName() == "ssoLogout") {
        if(m_localDebug) {
            qDebug() << "Code retour de reply : " << reply->error();
        }
        m_ssoAuthentifiedStatus = 0;
        disconnect(this, SIGNAL(finished(QNetworkReply*)),this, SLOT(slotFinishedSSOLogout(QNetworkReply*)));
    }
    emit ssoAuthStatus(m_ssoAuthentifiedStatus);

    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::slotFinishedSSOLogout (2)";
}

void AbulEduNetworkAccessManagerV1::slotSSOPopupClosed()
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1 : slotSSOPopupClosed (1) " << m_ssoPopupIsShow;

    m_ssoPopupIsShow = false;

    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1 : slotSSOPopupClosed (2) " << m_ssoPopupIsShow;
}

void AbulEduNetworkAccessManagerV1::slotLocalSocketMessageReceived(QString message)
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1::slotLocalSocketMessageReceived : " << message;

    //Si on est alacarte
    if(abeApp->applicationName() == "abuledu-alacarte") {
        //Contenu du message
        //HELLO:leterrier-calculment:ssoRequest
        QStringList params = message.split(":");
        if(m_localDebug)
            qDebug() << "AbulEduNetworkAccessManagerV1::slotLocalSocketMessageReceived " << params;
        //On se connecte maintenant sur le client
        if(params.count() > 1)
            m_localSocketClient->abeSlotChangeRemoteServer(params.at(1));
        QByteArray login(m_login.toUtf8());
        QByteArray pass(m_pass.toUtf8());
        m_localSocketClient->abeSlotSendMessageToServer("HELLO:ssoAnswer:" + login.toBase64() + ":" + pass.toBase64());
        //Reponse
        //HELLO:ssoAnswer:jgkeljgieajgiaejgera6432654
    }
    else {
        //Contenu du message
        QByteArray login = QByteArray::fromBase64(message.split(":").at(2).toAscii());
        QByteArray pass  = QByteArray::fromBase64(message.split(":").at(3).toAscii());
        if(m_localDebug)
            qDebug() << " ============================== Apres le Base64Decode : " << login << " et " << pass;

        //Si alacarte a retourné des choses intéressantes ...
        if(login != "") {
            setAbeLoginPass(login,pass);
            abeSSOLogin();
        }
    }
}

//Gestion de l'authentification proxy ...
void AbulEduNetworkAccessManagerV1::slotProxyAuthenticationRequired(QNetworkProxy proxy, QAuthenticator *authenticator)
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1 : slotProxyAuthenticationRequired." << m_proxyPassAsked;

    //On cherche voir si on a pas deja l'info sauvegardee
    QSettings settings(abeApp->organizationName(),"AbulEduProxyPasswordDialog");
    settings.beginGroup("proxy");

    //Si c'est la 1ere fois qu'on essaye ces parametres ... ok mais
    //sinon il faut la popup (dans ce cas c'est que les enregistrements sauvegardes ne sont plus bon)
    if((settings.value("server").toString() == this->proxy().hostName()) && (m_proxyPassAsked == 0)) {
        m_proxyLogin->clear();
        m_proxyLogin->append(settings.value("login").toString());
        m_proxyPass->clear();
        m_proxyPass->append(settings.value("pass").toString());
        QByteArray passc;
        passc.append(settings.value("pass").toString());
        authenticator->setUser(settings.value("login").toString());
        authenticator->setPassword(QByteArray::fromHex(QByteArray::fromBase64(passc)));
        QNetworkRequest redirection(m_networkReply->url());
        m_networkReply = this->get(redirection);
        //Detection loop
        m_proxyPassAsked++;
    }
    //et sinon direction popup
    else if(m_proxyPassAsked == 1) {
        AbulEduProxyPasswordDialog *dlg = new AbulEduProxyPasswordDialog(m_proxyLogin,m_proxyPass);
        dlg->setAbeLoginPassDialogNetworkReply(m_networkReply);
        dlg->setAbeLoginPassDialogAuthenticator(authenticator);
        dlg->show();
        m_proxyPassAsked++;
    }
}

void AbulEduNetworkAccessManagerV1::slotReplyError(QNetworkReply::NetworkError code)
{
    if(m_localDebug)
        qDebug() << "AbulEduNetworkAccessManagerV1 : slotReplyError." << code;

}

void AbulEduNetworkAccessManagerV1::slotSslErrors(QNetworkReply *reply, const QList<QSslError> &errors)
{
    if(m_localDebug) {
        qDebug() << "AbulEduNetworkAccessManagerV1 : slotSslErrors : " << errors;
        qDebug() << "AbulEduNetworkAccessManagerV1 : slotSslErrors : " << reply->errorString();
        qDebug() << "AbulEduNetworkAccessManagerV1 : slotSslErrors : " << reply->sslConfiguration().caCertificates();
    }
}
